/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM WebSourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: CollectorServer.java mcarpentier $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.tools.webadmin.datacollector.server;

import java.sql.Connection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hsqldb.Server;
import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerStates;
import org.ow2.petals.jmx.api.IPetalsAdminService.DomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.SubDomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.TopologyKeyNames;
import org.ow2.petals.jmx.exception.ConnectionErrorException;
import org.ow2.petals.tools.webadmin.datacollector.client.CollectorClient;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientException;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientInitializationException;
import org.ow2.petals.tools.webadmin.datacollector.dao.DAOException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorClientDoesNotExistException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerNotInitializedException;
import org.ow2.petals.tools.webadmin.datacollector.util.CollectorThread;
import org.ow2.petals.tools.webadmin.datacollector.util.DataCollectorConfigException;
import org.ow2.petals.tools.webadmin.datacollector.util.HibernateHelper;
import org.ow2.petals.tools.webadmin.datacollector.util.IHibernateHelper;
import org.ow2.petals.tools.webadmin.datacollector.util.Queuesmanager;
import org.ow2.petals.tools.webadmin.datacollector.util.ReconnectionThread;
import org.ow2.petals.tools.webadmin.datacollector.util.StringHelper;

public final class CollectorServer implements IHibernateHelper {

    private final Map<String, CollectorClient> collectorClients;

    private final Set<Map<String, String>> subdomainsProperties;

    private Map<String, String> domainProperties;

    private static Server hsqlServer = null;

    private CollectorThread collectorThread;

    private int queuesCollectingInterval = 10000;

    private int reconnectionInterval = 10000;

    private int reconnectionAttempts = 30;

    private long startTime = -1;

    private ReconnectionThread reconnectionThread;

    private final Queuesmanager qm;

    final static Logger log = Logger.getLogger(CollectorServer.class);

    /** properties given to the collector for initialization * */
    private final String host;

    private final String jndiPort;

    private final String user;

    private final String password;
    
    private static final String DATABASE_MODE = "database-mode";

    private static final String HSQL_DIRECTORY = "hsql-directory";

    private static final String DATABASE_HOST = "database-host";

    private static final String DATABASE_PORT = "database-port";

    private static final String DATABASE_NAME = "database-name";

    /**
     * 
     * @throws DataCollectorClientException
     * @throws DataCollectorClientInitializationException
     * @throws DataCollectorServerException
     *             SubDomainKeyNames
     */
    public CollectorServer(String host, String jndiPort, String user, String pwd, Properties props)
            throws DataCollectorServerException {

        if ("embedded".equals(props.getProperty(DATABASE_MODE))) {
            this.startDatabaseServer(props);
        }

        HibernateHelper.configureSession(props);

        this.qm = new Queuesmanager();
        this.host = host;
        this.jndiPort = jndiPort;
        this.password = pwd;
        this.user = user;
        this.startTime = (new Date()).getTime();
        this.collectorClients = new HashMap<String, CollectorClient>();
        this.subdomainsProperties = new HashSet<Map<String, String>>();
        this.initServer(false);
    }

    /**
     * Do all server init operations.<br>
     * First temp client will try to connect to petals and the retrieve all
     * connected PEtALS servers conf
     * 
     * @param firstClient
     * @throws DataCollectorServerException
     */
    public void initServer(boolean refresh) throws DataCollectorServerException {

        if (refresh) {
            for (final CollectorClient collectorClient : this.collectorClients.values()) {
                try {
                    if (collectorClient.getMonitoringService().isMonitoring()) {
                        collectorClient.getMonitoringService().stopListening();
                    }
                } catch (final DataCollectorClientException e) {
                    log.error(e);
                }
                try {
                    collectorClient.disconnect();
                } catch (final DataCollectorClientException e) {
                    log.error(e);
                }

            }
            this.collectorClients.clear();
            this.subdomainsProperties.clear();
        }

        try {
            final CollectorClient firstClient = new CollectorClient(this.host, Integer
                    .valueOf(this.jndiPort), this.user, this.password);
            final Set<Map<String, String>> topology = firstClient.getAdminService().getTopology();
            firstClient.disconnect();
            for (final Map<String, String> topologyElement : topology) {
                if (DomainKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {
                    this.domainProperties = topologyElement;
                } else if (SubDomainKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {
                    this.subdomainsProperties.add(topologyElement);
                } else if (ContainerKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {

                    // add only the started Container
                    if (ContainerStates.STATE_STARTED.equals(topologyElement
                            .get(ContainerKeyNames.CONF_STATE))) {
                        // switch the server view JMX RMI port to client
                        // view JMX JNDI port
                        topologyElement.put(ContainerKeyNames.CONF_JMX_JNDI_PORT, topologyElement
                                .remove(ContainerKeyNames.CONF_JMX_RMI_PORT));

                        final CollectorClient cc = new CollectorClient(topologyElement);

                        this.collectorClients.put(cc
                                .getServerProperty(ContainerKeyNames.CONF_CONTAINER_NAME), cc);
                        try {
                            // If monitoring for this server is activated, start
                            // listening for messages
                            if (cc.getMonitoringService().isMonitoring()) {
                                cc.getMonitoringService().startListening();
                            }

                        } catch (final DataCollectorClientException e) {
                            // The monitoring of this server is not
                            // available
                            log.debug(e);
                            cc.addContainerProperty(ContainerKeyNames.CONF_STATE,
                                    ContainerStates.STATE_STOPPED);
                        }
                    }
                }
            }

            // queues listening
            if (!refresh) {
                this.startQueuesListening();
            }

        } catch (final DataCollectorClientException e) {
            throw new DataCollectorServerException(e);
        } catch (final NumberFormatException e) {
            throw new DataCollectorServerException("Can't init JMX connection", e);
        }
    }

    /**
     * Check JDBC connection
     * 
     * @throws DataCollectorServerException
     */
    @SuppressWarnings("deprecation")
    public void checkDatabaseIsRunning() throws DataCollectorServerException {
        Session session = null;
        try {
            session = HibernateHelper.getSession();
            log.debug("Session on database OK : " + session.getStatistics().toString());
            final Connection con = session.connection();
            log.debug("Connection to database server OK : " + con.toString());
        } catch (final HibernateException e) {
            if (session != null) {
                session.close();
            }
            throw new DataCollectorServerException("Can't open connection to database server", e);
        } catch (final DataCollectorConfigException e) {
            if (session != null) {
                session.close();
            }
            throw new DataCollectorServerException("Can't configure the database server", e);
        }

    }

    public CollectorClient getCollectorClient(String serverName)
            throws DataCollectorServerException, DataCollectorClientDoesNotExistException,
            DataCollectorServerNotInitializedException {
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            if (this.collectorClients.containsKey(serverName)) {
                return this.collectorClients.get(serverName);
            } else {
                throw new DataCollectorClientDoesNotExistException("Can't find " + serverName
                        + " JMX client");
            }
        }
    }

    public Map<String, String> getDomainProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.domainProperties.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        }

        return this.domainProperties;
    }

    public Set<Map<String, String>> getSubdomainsProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.subdomainsProperties.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        }

        return this.subdomainsProperties;
    }

    public Set<Map<String, String>> getContainersProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.collectorClients.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            final Set<Map<String, String>> result = new HashSet<Map<String, String>>();
            for (final String client : this.collectorClients.keySet()) {
                result.add(this.collectorClients.get(client).getContainerProperties());
            }
            return result;
        }
    }

    public CollectorClient getDefaultCollectorClient() throws DataCollectorServerException,
            DataCollectorServerNotInitializedException {
        CollectorClient defaultClient = null;
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            defaultClient = this.collectorClients.values().iterator().next();
        }
        return defaultClient;

    }

    private void startDatabaseServer(Properties props) throws DataCollectorServerException {
        try {
            if (hsqlServer == null) {

                hsqlServer = (Server) Class.forName("org.hsqldb.Server").newInstance();

                final String databaseName = props.getProperty(DATABASE_NAME);
                final String databasePort = props.getProperty(DATABASE_PORT);
                final String databaseHost = props.getProperty(DATABASE_HOST);
                final String databaseDir = props.getProperty(HSQL_DIRECTORY);

                log.debug("server path read on configuration: " + databaseDir + "/" + databaseName);
                hsqlServer.setDatabasePath(0, databaseDir + "/" + databaseName);

                log.debug("server name read on configuration: " + databaseName);
                hsqlServer.setDatabaseName(0, databaseName);

                log.debug("server port read on configuration: " + databasePort);
                hsqlServer.setPort(Integer.parseInt(databasePort));

                log.debug("server host read on configuration: " + databaseHost);
                hsqlServer.setAddress(databaseHost);

                hsqlServer.setSilent(true);
                log.debug("silence activated on server");
                
                hsqlServer.setTrace(true);
                log.debug("trace wanted mode activated on server");
            }
            hsqlServer.checkRunning(true);
        } catch (final RuntimeException re) {
            log.debug("server state: " + hsqlServer.getState() + ", "
                    + hsqlServer.getStateDescriptor());
            log.debug("server address: " + hsqlServer.getAddress());
            log.info("hibernate server is stopped, trying to start it");
            hsqlServer.start();
            log.debug("server state after start: " + hsqlServer.getState() + ", "
                    + hsqlServer.getStateDescriptor());
            log.debug("server address after start: " + hsqlServer.getAddress());
        } catch (final InstantiationException e) {
            throw new DataCollectorServerException("Can't start database server", e);
        } catch (final IllegalAccessException e) {
            throw new DataCollectorServerException("Can't start database server", e);
        } catch (final ClassNotFoundException e) {
            throw new DataCollectorServerException("Can't start database server", e);
        }
    }

    public void startQueuesListening() {
        this.stopQueuesListening();
        // TODO : implement this in a proper way
        this.collectorThread = new CollectorThread(this);
        this.collectorThread.start();
    }

    public static void stopDatabaseServer() {
        if (hsqlServer != null) {
            Session session = null;
            try {
                session = HibernateHelper.getSession();
                // SQLQuery q = session.createSQLQuery("SHUTDOWN IMMEDIATELY");
                hsqlServer.stop();
                hsqlServer.shutdown();
            } catch (final Exception e) {
                log.error("Can't shutdown database", e);
            } finally {
                if (session != null) {
                    session.close();
                }
            }

        }
    }

    public void stopQueuesListening() {
        if ((this.collectorThread != null) && (this.collectorThread.isAlive())) {
            this.collectorThread.interrupt();
        }
    }

    public void stopMessagesListening() {
        try {
            if (this.collectorClients != null) {
                for (final CollectorClient c : this.collectorClients.values()) {
                    c.getMonitoringService().stopListening();
                }
            }
        } catch (final DataCollectorClientException e) {
            /** we just need to log the error, no more* */
            log.info("Couldn't remove message listener from PEtALS", e);
        }
    }

    public void collectAllQueues() {
        String currentClient = "";
        try {
            if (this.collectorClients != null) {
                for (final String client : this.collectorClients.keySet()) {
                    currentClient = client;
                    if (this.qm != null) {
                        this.qm.collectQueues(client, this.collectorClients.get(client)
                                .getTransportMonitoringService());
                    }
                }
            }
        } catch (final DataCollectorClientException e) {
            CollectorServer.log.error("Can't retrieve queues status", e);
            if (e.getCause() instanceof ConnectionErrorException) {
                /** check client connection timeout* */

                if ((StringHelper.isNotNullAndNotEmpty(currentClient))
                        && (this.collectorClients.containsKey(currentClient))) {
                    final CollectorClient cc = this.collectorClients.get(currentClient);
                    if (!cc.checkConnectionFailedCount()) {
                        this.collectorClients.remove(currentClient);
                        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
                            log
                                    .info("PEtALS connection seems to be lost, reconnection timer starting");
                            this.reconnectionThread = new ReconnectionThread(this);
                            this.reconnectionThread.start();
                        }
                    }
                }
            }
        } catch (final DAOException e) {
            CollectorServer.log.error("Can't retreive queues status", e);
        }

    }

    public int getQueuesCollectingInterval() {
        return this.queuesCollectingInterval;
    }

    public void setQueuesCollectingInterval(int freq) {
        this.queuesCollectingInterval = freq;
        if (this.collectorThread != null) {
            this.collectorThread.setRefreshFrequency(freq);
        }
    }

    public int getReconnectionInterval() {
        return this.reconnectionInterval;
    }

    public void setReconnectionInterval(int reconnectionInterval) {
        this.reconnectionInterval = reconnectionInterval;
        if (this.reconnectionThread != null) {
            this.reconnectionThread.setRefreshFrequency(reconnectionInterval);
        }
    }

    public int getReconnectionAttempts() {
        return this.reconnectionAttempts;
    }

    public void setReconnectionAttempts(int reconnectionAttempts) {
        this.reconnectionAttempts = reconnectionAttempts;
        if (this.reconnectionThread != null) {
            this.reconnectionThread.setMaxAttempts(reconnectionAttempts);
        }
    }

    public String getDatabasePath() {
        if (hsqlServer != null) {
            return hsqlServer.getDatabasePath(0, true);
        } else {
            return "unknown";
        }
    }

    public int getDatabasePort() {
        if (hsqlServer != null) {
            return hsqlServer.getPort();
        } else {
            return -1;
        }
    }

    public long getStartTime() {
        return this.startTime;
    }

    public String getDatabaseVersion() {
        if (hsqlServer != null) {
            return hsqlServer.getProductName() + " " + hsqlServer.getProductVersion();
        } else {
            return "unknown";
        }
    }

}
